"""Logging and log related utilities."""

from __future__ import annotations

from .formatters import (
    console_log_formatter,
    json_formatter,
    key_value_formatter,
    plain_formatter,
)
from .job_logging_service import (
    JobLoggingService,
    MissingJobLogException,
    SizeThresholdJobLogException,
)
from .output_logger import OutputLogger
from .utils import (
    DEFAULT_LEVEL,
    LEVELS,
    LogFormat,
    capture_subprocess_output,
    setup_logging,
)

__all__ = [
    "DEFAULT_LEVEL",
    "LEVELS",
    "JobLoggingService",
    "LogFormat",
    "MissingJobLogException",
    "OutputLogger",
    "OutputLogger",
    "SizeThresholdJobLogException",
    "capture_subprocess_output",
    "console_log_formatter",
    "json_formatter",
    "key_value_formatter",
    "plain_formatter",
    "setup_logging",
]



"""Storage managers for job state."""

from __future__ import annotations

import dataclasses
import json
import typing as t
from abc import ABC, abstractmethod
from contextlib import contextmanager

from meltano.core.utils import merge

if t.TYPE_CHECKING:
    from collections.abc import Generator, Iterable
    from io import TextIOWrapper


class UnsupportedStateBackendURIError(Exception):
    """Provided state backend URI is not supported."""


class MissingStateBackendSettingsError(Exception):
    """Required setting values for a configured State Backend are missing."""


class StateIDLockedError(Exception):
    """A job attempted to acquire a lock on an already-locked state ID."""


@dataclasses.dataclass
class MeltanoState:
    """State object."""

    state_id: str
    partial_state: dict[str, t.Any] | None = None
    completed_state: dict[str, t.Any] | None = None

    @classmethod
    def from_json(cls, state_id: str, json_str: str) -> MeltanoState:
        """Create MeltanoState from json representation.

        Args:
            state_id: the state ID of the job to create MeltanoState for.
            json_str: the json representation of state to use.

        Returns:
            MeltanoState representing args.
        """
        state_dict = json.loads(json_str)
        return cls(
            state_id=state_id,
            completed_state=state_dict.get("completed", {}),
            partial_state=state_dict.get("partial", {}),
        )

    @classmethod
    def from_file(cls, state_id: str, file_obj: TextIOWrapper) -> MeltanoState:
        """Create MeltanoState from a file-like object containing state json.

        Args:
            state_id: the state_id for the MeltanoState
            file_obj: the file-like object containing state json.

        Returns:
            MeltanoState
        """
        return cls.from_json(state_id=state_id, json_str=file_obj.read())

    def json(self) -> str:
        """Get the json representation of this MeltanoState.

        Returns:
            json representation of this MeltanoState
        """
        return json.dumps(
            {"completed": self.completed_state, "partial": self.partial_state},
        )

    def json_merged(self) -> str:
        """Return the json representation of partial state merged onto complete state.

        Returns:
            json representation of partial state merged onto complete state.
        """
        return json.dumps(merge(self.partial_state, self.completed_state))

    def merge_partial(self, state: MeltanoState) -> None:
        """Merge provided partial state onto this MeltanoState.

        Args:
            state: the partial state to merge onto this state.
        """
        self.partial_state = merge(
            state.partial_state,
            self.partial_state,
        )

    def is_complete(self) -> bool:
        """Check if this MeltanoState is complete.

        Returns:
            True if complete, else False
        """
        return bool(self.completed_state)


class StateStoreManager(ABC):
    """Base state store manager."""

    def __init__(self, **kwargs: t.Any) -> None:  # noqa: B027
        """Initialize state store manager.

        Args:
            kwargs: additional keyword arguments
        """

    @property
    @abstractmethod
    def label(self) -> str:
        """Get the label for the StateStoreManager."""
        ...

    @abstractmethod
    def set(self, state: MeltanoState) -> None:
        """Set the job state for the given state_id.

        Args:
            state: the state to set.

        Raises:
            NotImplementedError: always, this is an abstract method
        """
        ...

    @abstractmethod
    def get(self, state_id: str) -> MeltanoState | None:
        """Get the job state for the given state_id.

        Args:
            state_id: the name of the job to get state for.

        Raises:
            NotImplementedError: always, this is an abstract method
        """
        ...

    @abstractmethod
    def clear(self, state_id: str) -> None:
        """Clear state for the given state_id.

        Args:
            state_id: the state_id to clear state for
        """
        ...

    def clear_all(self) -> int:
        """Clear all states.

        Override this method if the store supports bulk deletion.

        Returns:
            The number of states cleared from the store.
        """
        count = 0
        for state_id in self.get_state_ids("*"):
            self.clear(state_id)
            count += 1
        return count

    @abstractmethod
    def get_state_ids(self, pattern: str | None = None) -> Iterable[str]:
        """Get all state_ids available in this state store manager.

        Args:
            pattern: glob-style pattern to filter by
        """
        ...

    @abstractmethod
    @contextmanager
    def acquire_lock(
        self,
        state_id: str,
        *,
        retry_seconds: int,
    ) -> Generator[None, None, None]:
        """Acquire a naive lock for the given job's state.

        Args:
            state_id: the state_id to lock
            retry_seconds: the number of seconds to wait before retrying
        """
        ...



from __future__ import annotations  # noqa: D100

import typing as t

if t.TYPE_CHECKING:
    from collections.abc import Callable


class visit_with:  # noqa: D101, N801
    def __init__(self, visit: Callable):  # noqa: D107
        self.visit = visit

    def __call__(self, base_cls):  # noqa: ANN001, ANN204, D102
        class Visitor(base_cls):
            def visit(inner_self, node, *args, **kwargs):  # noqa: ANN001, ANN002, ANN003, ANN202, N805
                return self.visit(node, inner_self, *args, **kwargs)

        return Visitor



from __future__ import annotations

import importlib.resources

from meltano.core.tracking import schemas

schemas_dir = (
    importlib.resources.files("meltano.core.tracking")
    / "iglu-client-embedded"
    / "schemas"
    / "com.meltano"
)

versions_in_use = {
    x.name: x.version.split("-")
    for x in vars(schemas).values()
    if isinstance(x, schemas.IgluSchema)
}

versions_available = {
    schema_dir.name: max(
        x.name.split("-") for x in (schema_dir / "jsonschema").iterdir()
    )
    for schema_dir in schemas_dir.iterdir()
}


def test_using_latest_schemas() -> None:
    assert versions_in_use == versions_available



"""Various utils and formatters for log rendering control."""

from __future__ import annotations

import sys
import typing as t

import click
import structlog
import structlog.typing
from rich.console import Console
from rich.traceback import Traceback, install

from meltano.core.utils import get_no_color_flag

if sys.version_info < (3, 11):
    from typing_extensions import Unpack
else:
    from typing import Unpack  # noqa: ICN003

if t.TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from types import TracebackType

    from structlog.types import Processor

install(suppress=[click])

TIMESTAMPER = structlog.processors.TimeStamper(fmt="iso")

LEVELED_TIMESTAMPED_PRE_CHAIN: t.Sequence[Processor] = (
    # Add the log level and a timestamp to the event_dict if the log entry
    # is not from structlog.
    structlog.stdlib.add_log_level,
    TIMESTAMPER,
)


class LoggingFeatures(t.TypedDict, total=False):
    """Logging features that can be enabled in a formatter."""

    callsite_parameters: bool
    """Enable filename, line number, and function name in log entries.

    https://www.structlog.org/en/stable/api.html#structlog.processors.CallsiteParameter.
    """

    dict_tracebacks: bool
    """Enable tracebacks dictionaries in log entries.

    https://www.structlog.org/en/stable/api.html#structlog.processors.dict_tracebacks.
    """

    show_locals: bool
    """Whether to show local variables in the traceback.

    https://www.structlog.org/en/stable/api.html#module-structlog.tracebacks.
    """


# Convert boolean kwargs to LoggingFeatures enum.
def _processors_from_kwargs(
    **features: Unpack[LoggingFeatures],
) -> t.Generator[Processor, None, None]:
    if features.get("callsite_parameters", False):
        yield structlog.processors.CallsiteParameterAdder(
            parameters=(
                structlog.processors.CallsiteParameter.PATHNAME,
                structlog.processors.CallsiteParameter.LINENO,
                structlog.processors.CallsiteParameter.FUNC_NAME,
            ),
        )

    if features.get("dict_tracebacks", False):
        show_locals = features.get("show_locals", False)
        yield structlog.processors.ExceptionRenderer(
            structlog.tracebacks.ExceptionDictTransformer(show_locals=show_locals),
        )


def rich_exception_formatter_factory(
    color_system: t.Literal["auto", "standard", "256", "truecolor", "windows"] = "auto",
    *,
    no_color: bool | None = None,
    show_locals: bool = False,
) -> Callable[[t.TextIO, structlog.types.ExcInfo], None]:
    """Create an exception formatter for logging using the rich package.

    Examples:
    >>> rich_traceback = rich_exception_formatter_factory(color_system="truecolor")
    >>> plane_rich_traceback = rich_exception_formatter_factory(no_color=True)

    Args:
        color_system: The color system supported by your terminal.
        no_color: Enabled no color mode, or None to auto detect. Defaults to None.
        show_locals: Whether to show local variables in the traceback.

    Returns:
        Exception formatter function.
    """

    def _traceback(
        sio,  # noqa: ANN001
        exc_info: tuple[type[t.Any], BaseException, TracebackType | None],
    ) -> None:
        sio.write("\n")
        Console(file=sio, color_system=color_system, no_color=no_color).print(
            Traceback.from_exception(
                *exc_info,
                show_locals=show_locals,
            ),
        )

    return _traceback


def _process_formatter(
    *processors: Processor,
    **kwargs: t.Any,
) -> structlog.stdlib.ProcessorFormatter:
    """Use _process_formatter to configure a structlog.stdlib.ProcessFormatter.

    It will automatically add log level and timestamp fields to any log entries
    not originating from structlog.

    Args:
        *processors: One or more structlog message processors such as
            `structlog.dev.ConsoleRenderer`.
        **kwargs: Additional keyword arguments to pass to the logging.Formatter
            constructor.

    Returns:
        A log record formatter.
    """
    return structlog.stdlib.ProcessorFormatter(
        processors=processors,
        foreign_pre_chain=LEVELED_TIMESTAMPED_PRE_CHAIN,
        **kwargs,
    )


def console_log_formatter(
    *,
    colors: bool = False,
    callsite_parameters: bool = False,
    show_locals: bool = False,
) -> structlog.stdlib.ProcessorFormatter:
    """Create a logging formatter for console rendering that supports colorization.

    Args:
        colors: Add color to output.
        callsite_parameters: Whether to include callsite parameters in the output.
        show_locals: Whether to show local variables in the traceback.

    Returns:
        A configured console log formatter.
    """
    colors = colors and not get_no_color_flag()

    if colors:
        exception_formatter = rich_exception_formatter_factory(
            color_system="truecolor",
            show_locals=show_locals,
        )
    else:
        exception_formatter = rich_exception_formatter_factory(
            no_color=True,
            show_locals=show_locals,
        )

    return _process_formatter(
        *_processors_from_kwargs(callsite_parameters=callsite_parameters),
        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
        structlog.dev.ConsoleRenderer(
            colors=colors,
            exception_formatter=exception_formatter,
        ),
    )


def key_value_formatter(
    *,
    sort_keys: bool = False,
    key_order: Sequence[str] | None = None,
    drop_missing: bool = False,
    callsite_parameters: bool = False,
) -> structlog.stdlib.ProcessorFormatter:
    """Create a logging formatter that renders lines in key=value format.

    Args:
        sort_keys: Whether to sort keys when formatting.
        key_order: List of keys that should be rendered in this exact order.
            Missing keys will be rendered as None, extra keys depending on
            *sort_keys* and the dict class.
        drop_missing: When True, extra keys in *key_order* will be dropped
            rather than rendered as None.
        callsite_parameters: Whether to include callsite parameters in the output.

    Returns:
        A configured key=value formatter.
    """
    return _process_formatter(
        *_processors_from_kwargs(callsite_parameters=callsite_parameters),
        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
        structlog.processors.KeyValueRenderer(
            sort_keys=sort_keys,
            key_order=key_order,
            drop_missing=drop_missing,
        ),
    )


def json_formatter(
    *,
    callsite_parameters: bool = False,
    dict_tracebacks: bool = True,
    show_locals: bool = False,
) -> structlog.stdlib.ProcessorFormatter:
    """Create a logging formatter that renders lines in JSON format.

    Args:
        callsite_parameters: Whether to include callsite parameters in the JSON output.
        dict_tracebacks: Whether to include tracebacks in the JSON output.
        show_locals: Whether to include local variables in the traceback.

    Returns:
        A configured JSON formatter.
    """
    return _process_formatter(
        *_processors_from_kwargs(
            callsite_parameters=callsite_parameters,
            dict_tracebacks=dict_tracebacks,
            show_locals=show_locals,
        ),
        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
        structlog.processors.JSONRenderer(),
    )


def _event_renderer(
    logger: structlog.typing.WrappedLogger,  # noqa: ARG001
    name: str,  # noqa: ARG001
    event_dict: structlog.typing.EventDict,
) -> str | bytes:
    """Render an event dictionary as a string.

    Args:
        logger: The logger instance.
        name: The logger name.
        event_dict: The event dictionary.

    Returns:
        The rendered event dictionary.
    """
    return event_dict["event"]


def plain_formatter(
    *,
    fmt: str | None = None,
    datefmt: str | None = None,
    style: str = "%",
    validate: bool = True,
) -> structlog.stdlib.ProcessorFormatter:
    """Create a logging formatter that renders lines in a simple format.

    Args:
        fmt: The format string.
        datefmt: The date format string.
        style: The format style.
        validate: Whether to validate the format string.
        features: Logging features to enable.

    Returns:
        A configured simple formatter.
    """
    return _process_formatter(
        structlog.stdlib.filter_by_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        structlog.processors.UnicodeDecoder(),
        _event_renderer,
        fmt=fmt,
        datefmt=datefmt,
        style=style,
        validate=validate,
    )



from __future__ import annotations

import asyncio
import datetime
import logging
import zoneinfo

import pytest
import time_machine

from meltano.core.logging.utils import (
    LogFormat,
    capture_subprocess_output,
    default_config,
)


class AsyncReader(asyncio.StreamReader):
    def __init__(self, lines: list[bytes]):
        self.lines = lines

    def at_eof(self) -> bool:
        return not self.lines

    async def readline(self) -> bytes:
        return self.lines.pop(0) if self.lines else b""


@pytest.mark.asyncio
async def test_capture_subprocess_output() -> None:
    input_lines = [b"LINE\n", b"LINE 2\n", b"\xed\n"]
    output_lines = []

    class LineWriter:
        def writeline(self, line: str) -> None:
            output_lines.append(line)

    reader = AsyncReader(input_lines)

    await capture_subprocess_output(reader, LineWriter())
    assert output_lines == ["LINE\n", "LINE 2\n", "�\n"]


@pytest.mark.parametrize(
    ("log_format", "expected"),
    (
        pytest.param(
            LogFormat.colored,
            "\x1b[2m2021-01-01T00:00:00Z\x1b[0m [\x1b[32m\x1b[1minfo     \x1b[0m] \x1b[1mtest                          \x1b[0m",  # noqa: E501
            id="colored",
        ),
        pytest.param(
            LogFormat.uncolored,
            "2021-01-01T00:00:00Z [info     ] test",
            id="uncolored",
        ),
        pytest.param(
            LogFormat.json,
            '{"event": "test", "level": "info", "timestamp": "2021-01-01T00:00:00Z"}',
            id="json",
        ),
        pytest.param(
            LogFormat.key_value,
            "timestamp='2021-01-01T00:00:00Z' level='info' event='test' logger=None",
            id="key_value",
        ),
        pytest.param(
            LogFormat.plain,
            "test",
            id="plain",
        ),
    ),
)
def test_default_logging_config_format(
    log_format: LogFormat,
    expected: str,
    monkeypatch: pytest.MonkeyPatch,
):
    config = default_config("info", log_format=log_format)
    assert log_format in config["formatters"]
    assert config["handlers"]["console"]["formatter"] == log_format
    assert config["loggers"][""]["level"] == "INFO"

    # Create a logger that uses the config
    formatter_config = config["formatters"][log_format]
    formatter_class = formatter_config.pop("()")
    formatter = formatter_class(**formatter_config)
    formatter.logger = logging.getLogger("test")  # noqa: TID251
    formatter.logger.setLevel(logging.DEBUG)

    # Mock current time and process ID
    with (
        time_machine.travel(
            datetime.datetime(2021, 1, 1, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("UTC")),
            tick=False,
        ),
        monkeypatch.context() as m,
    ):
        m.setattr("os.getpid", lambda: 95803)
        record = logging.LogRecord(
            name="test_logger",
            level=logging.INFO,
            pathname="",
            lineno=0,
            msg="test",
            args=(),
            exc_info=None,
        )

        # Test the formatted message
        formatted = formatter.format(record)
        assert formatted == expected



from __future__ import annotations

import json
import logging
import re
import typing as t
from types import TracebackType

import pytest
import structlog.exceptions

from meltano.core.logging import formatters

if t.TYPE_CHECKING:
    from pathlib import Path

ANSI_RE = re.compile(r"\033\[[;?0-9]*[a-zA-Z]")


class FakeCode:
    """A fake code object that can be used to test log formatters."""

    def __init__(self, co_filename, co_name) -> None:
        self.co_filename = co_filename
        self.co_name = co_name

    def co_positions(self):
        yield 0, 10, 0, 40


class FakeFrame:
    """A fake traceback frame that can be used to test log formatters."""

    def __init__(self, f_code, f_globals, f_locals=None) -> None:
        self.f_code = f_code
        self.f_globals = f_globals
        self.f_locals = f_locals or {}
        self.f_lasti = 0


class FakeTraceback:  # pragma: no cover
    """A fake traceback that can be used to test log formatters."""

    def __init__(self, frames, line_nums) -> None:
        if len(frames) != len(line_nums):
            raise ValueError("Ya messed up!")  # noqa: EM101
        self._frames = frames
        self._line_nums = line_nums
        self.tb_frame = frames[0]
        self.tb_lineno = line_nums[0]

    @property
    def tb_next(self):
        if len(self._frames) > 1:  # noqa: RET503
            return FakeTraceback(self._frames[1:], self._line_nums[1:])


class TestLogFormatters:
    @pytest.fixture
    def record(self):
        return logging.LogRecord(
            name="test",
            level=logging.INFO,
            pathname="/path/to/my_module.py",
            lineno=1,
            msg="test",
            args=None,
            exc_info=None,
            func="my_func",
        )

    @pytest.fixture
    def record_with_exception(self, tmp_path: Path):
        path = tmp_path / "my_module.py"
        path.write_text("cause_an_error()\n")

        tb = t.cast(
            TracebackType,
            FakeTraceback(
                [
                    FakeFrame(
                        FakeCode(str(path), "test"),
                        {"__name__": "__main__"},
                        {"my_var": "my_value"},
                    ),
                ],
                [1],
            ),
        )

        return logging.LogRecord(
            name="test",
            level=logging.DEBUG,
            pathname="test",
            lineno=1,
            msg="test",
            args=None,
            exc_info=(
                Exception,
                Exception("test"),
                tb,
            ),
        )

    def test_console_log_formatter_colors(self, record, monkeypatch) -> None:
        monkeypatch.delenv("NO_COLOR", raising=False)
        formatter = formatters.console_log_formatter(colors=True)
        assert ANSI_RE.match(formatter.format(record))

    def test_console_log_formatter_no_colors(self, record) -> None:
        formatter = formatters.console_log_formatter(colors=False)
        assert not ANSI_RE.match(formatter.format(record))

    def test_console_log_formatter_no_locals(self, record_with_exception) -> None:
        formatter = formatters.console_log_formatter(show_locals=False)
        output = formatter.format(record_with_exception)
        assert "locals" not in output
        assert "my_var = 'my_value'" not in output

    def test_console_log_formatter_show_locals(self, record_with_exception) -> None:
        formatter = formatters.console_log_formatter(show_locals=True)
        output = formatter.format(record_with_exception)
        assert "locals" in output
        assert "my_var = 'my_value'" in output

    def test_key_value_formatter(self, record):
        formatter = formatters.key_value_formatter()
        output = formatter.format(record)
        assert "event='test' level='info'" in output
        assert "pathname='/path/to/my_module.py'" not in output

        formatter = formatters.key_value_formatter(callsite_parameters=True)
        output = formatter.format(record)
        assert "pathname='/path/to/my_module.py'" in output
        assert "lineno=1" in output
        assert "func_name='my_func'" in output

    def test_json_formatter_callsite_parameters(self, record):
        formatter = formatters.json_formatter(callsite_parameters=True)
        output = formatter.format(record)
        message_dict = json.loads(output)
        assert message_dict["pathname"] == "/path/to/my_module.py"
        assert message_dict["lineno"] == 1
        assert message_dict["func_name"] == "my_func"

    def test_json_formatter_exception(self, record_with_exception) -> None:
        formatter = formatters.json_formatter()
        output = formatter.format(record_with_exception)
        message_dict = json.loads(output)

        assert "exception" in message_dict

        exception_list = message_dict["exception"]
        assert isinstance(exception_list, list)
        assert len(exception_list) == 1
        assert exception_list[0]["exc_type"] == "Exception"
        assert exception_list[0]["exc_value"] == "test"

        formatter = formatters.json_formatter(dict_tracebacks=False)
        output = formatter.format(record_with_exception)
        message_dict = json.loads(output)

        assert "exception" not in message_dict

    def test_json_formatter_locals(self, record_with_exception) -> None:
        formatter = formatters.json_formatter(show_locals=True)
        output = formatter.format(record_with_exception)
        message_dict = json.loads(output)
        assert "locals" in message_dict["exception"][0]["frames"][0]

        formatter = formatters.json_formatter(show_locals=False)
        output = formatter.format(record_with_exception)
        message_dict = json.loads(output)
        assert "locals" not in message_dict["exception"][0]["frames"][0]

    def test_plain_formatter(self, record) -> None:
        formatter = formatters.plain_formatter(fmt="%(levelname)s %(name)s")
        formatter.logger = logging.getLogger("test")  # noqa: TID251
        formatter.logger.setLevel(logging.INFO)
        output = formatter.format(record)
        assert output == "INFO test"

    def test_plain_formatter_drop_event(self, record) -> None:
        formatter = formatters.plain_formatter(fmt="%(levelname)s %(name)s")
        formatter.logger = logging.getLogger("test")  # noqa: TID251
        formatter.logger.setLevel(logging.WARNING)

        with pytest.raises(structlog.exceptions.DropEvent):
            formatter.format(record)



"""Interactivity utils."""

from __future__ import annotations

import enum
import sys

if sys.version_info < (3, 11):
    from backports.strenum import StrEnum
else:
    from enum import StrEnum


class InteractionStatus(StrEnum):
    """Interaction status constants."""

    EXIT = enum.auto()
    SKIP = enum.auto()
    START = enum.auto()
    RETRY = enum.auto()



"""Validation command."""

from __future__ import annotations

import shutil
import sys
import typing as t

import click
import structlog

from meltano.cli.params import InstallPlugins, get_install_options, pass_project
from meltano.cli.utils import (
    CliEnvironmentBehavior,
    InstrumentedCmd,
    propagate_stop_signals,
)
from meltano.core.db import project_engine
from meltano.core.plugin_install_service import PluginInstallReason
from meltano.core.utils import run_async
from meltano.core.validation_service import ValidationOutcome, ValidationsRunner

if t.TYPE_CHECKING:
    from collections import abc

    from sqlalchemy.orm.session import Session

    from meltano.core.project import Project

logger = structlog.getLogger(__name__)

TEST_LINE_LENGTH = 60

install, no_install, only_install = get_install_options(include_only_install=True)


def write_sep_line(title: str, sepchar: str, **kwargs) -> None:  # noqa: ANN003
    """Write a separator line in the terminal."""
    terminal_width, _ = shutil.get_terminal_size()
    char_count = (
        terminal_width - len(title) - 2  # Whitespace between sepchar and title
    ) // (2 * len(sepchar))
    char_count = max(char_count, 1)

    fill = sepchar * char_count
    line = f"\n{fill} {title} {fill}"

    if len(line) + len(sepchar.rstrip()) <= terminal_width:
        line += sepchar.rstrip()

    click.secho(line, **kwargs)


class CommandLineRunner(ValidationsRunner):
    """Validator that runs in the CLI."""

    async def run_test(self, name: str) -> int:  # type: ignore[override]
        """Run a test command.

        Args:
            name: Test command name to invoke.

        Returns:
            Exit code for the plugin invocation.
        """
        write_sep_line(f"{self.plugin_name}:{name}", "=", bold=True)
        handle = await self.invoker.invoke_async(command=name)
        with propagate_stop_signals(handle):
            return await handle.wait()


@click.command(
    cls=InstrumentedCmd,
    short_help="Run validations using plugins' tests.",
    environment_behavior=CliEnvironmentBehavior.environment_optional_use_default,
)
@click.option(
    "--all",
    "all_tests",
    is_flag=True,
    help="Run all defined tests in all plugins.",
)
@click.argument(
    "plugin_tests",
    default=None,
    required=False,
    nargs=-1,
)
@install
@no_install
@only_install
@pass_project(migrate=True)
@run_async
async def test(
    project: Project,
    *,
    all_tests: bool,
    install_plugins: InstallPlugins,
    plugin_tests: tuple[str, ...] = (),
) -> None:
    """Run validations using plugins' tests.

    \b
    Read more at https://docs.meltano.com/reference/command-line-interface#test
    """  # noqa: D301
    _, session_maker = project_engine(project)
    session = session_maker()

    collected = CommandLineRunner.collect(project, select_all=all_tests)

    for plugin_test in plugin_tests:
        try:
            plugin_name, test_name = plugin_test.split(":")
        except ValueError:
            plugin_name, test_name = plugin_test, None

        if test_name:
            collected[plugin_name].select_test(test_name)
        else:
            collected[plugin_name].select_all()

    await install_plugins(
        project,
        [c.invoker.plugin for c in collected.values() if c.tests_selection],
        reason=PluginInstallReason.AUTO,
    )

    exit_codes = await _run_plugin_tests(session, collected.values())
    click.echo()
    _report_and_exit(exit_codes)


async def _run_plugin_tests(
    session: Session,
    runners: abc.Iterable[ValidationsRunner],
) -> dict[str, dict[str, int]]:
    return {runner.plugin_name: await runner.run_all(session) for runner in runners}


def _report_and_exit(results: dict[str, dict[str, int]]) -> None:
    exit_code = 0
    failed_count = 0
    passed_count = 0

    for plugin_name, test_results in results.items():
        for test_name, test_exit_code in test_results.items():
            outcome = ValidationOutcome.from_exit_code(test_exit_code)
            if test_exit_code == 0:
                passed_count += 1
            else:
                exit_code = 1
                failed_count += 1
            styled_result = click.style(outcome, fg=outcome.color)
            plugin_test = click.style(f"{plugin_name}:{test_name}", bold=True)
            click.echo(f"{plugin_test.ljust(TEST_LINE_LENGTH, '.')} {styled_result}")

    status = "successfully" if failed_count == 0 else "with failures"
    message = (
        f"Testing completed {status}. "
        f"{passed_count} test(s) successful. {failed_count} test(s) failed."
    )

    write_sep_line(message, "=", fg=("red" if exit_code else "green"))

    sys.exit(exit_code)
